/*	$Id: PaletteGik.cc 2008/04/07 17:29:23 mallet $ */

/*
 * Copyright (c) 2007 LAAS/CNRS                       --  Mon Apr 16 2007
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include "tasks/hppGikWholeBodyTask.h"
#include "tasks/hppGikHandTask.h"
#include "tasks/hppGikReachTask.h"
#include "tasks/hppGikHalfSittingTask.h"
#include "tasks/hppGikMetaTask.h"
#include "constraints/hppGikComConstraint.h"
#include "motionplanners/elements/hppGikInterpolatedElement.h"
#include "hppGikTools.h"

#include "PaletteGik.h"
#include "HRP2DynamicRobot.h"

#include "hrp2Dynamics/hrp2OptHumanoidDynamicRobot.h"

#include "hrp2_14/hrp2_14.h"

#define M4_IJ MAL_S4x4_MATRIX_ACCESS_I_J

using namespace boost::numeric::ublas;

PaletteGik::PaletteGik ( const char *path, const char *model ) : attSamplingPeriod ( 10e-3 )
{
    attHRP2MotionName = "GIK_motion";

    createStandingRobot ( path, model );

    attLastMotion = new ChppRobotMotion ( robot, 0.0, attSamplingPeriod );
    attGikMotion = new ChppRobotMotion ( robot, 0.0, attSamplingPeriod );

    attGikSolver = new ChppGikSolver ( *attStandingRobot->robot() );


    vectorN activated =  attStandingRobot->maskFactory()->upperBodyMask();
    vectorN weights = attStandingRobot->maskFactory()->weightsDoubleSupport();
    vectorN combined = weights;
    for ( unsigned int i=0;i<combined.size();i++ )
        combined ( i ) *= activated ( i );
    for ( unsigned int i=0;i<6;i++ )
        combined ( i ) = 0.0;

    /*
    vectorN combined =  attStandingRobot->maskFactory()->upperBodyMask();
    for (unsigned int i=0;i<6;i++)
        combined(i) = 0.0;
    */

    attGikSolver->weights ( combined );

    attSittingTasksDuration = 5.0;
}

PaletteGik::~PaletteGik()
{
    delete attLastMotion;
    delete attStandingRobot;
    delete attGikSolver;
}

void PaletteGik::createStandingRobot ( const char *ipath, const char *imodel )
{
    CjrlRobotDynamicsObjectConstructor<
    dynamicsJRLJapan::DynamicMultiBody,
    Chrp2OptHumanoidDynamicRobot,
    dynamicsJRLJapan::JointFreeflyer,
    dynamicsJRLJapan::JointRotation,
    dynamicsJRLJapan::JointTranslation,
    dynamicsJRLJapan::Body> jrlRobotFactory;

    robot = jrlRobotFactory.createhumanoidDynamicRobot();

    dynamicsJRLJapan::HumanoidDynamicMultiBody *aHDMB;
    aHDMB = ( dynamicsJRLJapan::HumanoidDynamicMultiBody* ) robot;

    std::string path = ipath;
    std::string model = imodel;
    std::string specfile = HRP2_14_SPECFILE;

    aHDMB->parserVRML ( path + "/", model, HRP2_14_RANKFILE );
    aHDMB->SetHumanoidSpecificitiesFile ( specfile );
    aHDMB->SetTimeStep ( attSamplingPeriod );
    aHDMB->setComputeVelocity ( true );
    aHDMB->setComputeMomentum ( true );
    aHDMB->setComputeCoM ( true );
    aHDMB->setComputeAcceleration ( true );
    aHDMB->setComputeZMP ( true );

    aHDMB->setComputeSkewCoM ( false );
    aHDMB->setComputeAccelerationCoM ( false );
    aHDMB->setComputeBackwardDynamics ( false );


    unsigned int nDof = robot->numberDof();
    vectorN halfsittingConf ( nDof );

    //Half sitting
    double dInitPos[40] = HALFSITTINGPOSITION_DEG_OPENHRP;

    //waist x y z
    halfsittingConf ( 0 ) = 0.0;
    halfsittingConf ( 1 ) = 0.0;
    halfsittingConf ( 2 ) = 0.6487;
    //waist roll pitch yaw
    halfsittingConf ( 3 ) = 0.0;
    halfsittingConf ( 4 ) = 0.0;
    halfsittingConf ( 5 ) = 0.0;

    //joints
    for ( unsigned int i=6;i<nDof;i++ )
        halfsittingConf ( i ) = dInitPos[i-6]*M_PI/180;

    /* apply initial static state*/
    zero_vector<double> zeros ( robot->numberDof() );
    robot->currentConfiguration ( halfsittingConf );
    robot->currentVelocity ( zeros );
    robot->currentAcceleration ( zeros );
    robot->computeForwardKinematics();

    //set gaze origin and direction
    vector3d gazeDir,gazeOrigin;
    gazeDir[0] = 1;
    gazeDir[1] = 0;
    gazeDir[2] = 0;

    gazeOrigin[0] = 0.01;
    gazeOrigin[1] = 0;
    gazeOrigin[2] = 0.145;

    aHDMB->gaze ( ( const vector3d& ) gazeDir, ( const vector3d& ) gazeOrigin );

    attCurrentStartConfig = halfsittingConf;
    attNextStartConfig = halfsittingConf;

    attStandingRobot = new ChppGikStandingRobot ( robot );
    attStandingRobot->staticState ( halfsittingConf );
}

void  PaletteGik::reset()
{
    attStandingRobot->staticHalfsitting();
    attCurrentStartConfig = attStandingRobot->halfsittingConfiguration();
    attNextStartConfig = attCurrentStartConfig;
}

void PaletteGik::setConfiguration ( vectorN& inConfiguration )
{
    attStandingRobot->staticState ( inConfiguration );

    std::cout << robot->leftFoot()->currentTransformation() << "\n";
    std::cout << robot->rightFoot()->currentTransformation() << "\n";

    double rz = M4_IJ ( robot->rightFoot()->currentTransformation(),2,3 );
    double lz = M4_IJ ( robot->leftFoot()->currentTransformation(),2,3 );
    double meanAnkleZ = ( rz+lz ) /2;

    vectorN config = inConfiguration;

    config ( 2 ) = config ( 2 ) - ( meanAnkleZ - robot->footHeight() );

    attStandingRobot->staticState ( config );

    std::cout << robot->leftFoot()->currentTransformation() << "\n";
    std::cout << robot->rightFoot()->currentTransformation() << "\n";

    attCurrentStartConfig = robot->currentConfiguration();
    attNextStartConfig = attCurrentStartConfig;
}

void PaletteGik::sittingSetConfiguration ( const vectorN& inConfiguration )
{
    vectorN conf = inConfiguration;

    for ( unsigned int i=0;i<6;i++ )
        conf ( i ) = 0.0;
    attStandingRobot->staticState ( inConfiguration );

    attCurrentStartConfig = robot->currentConfiguration();
    attNextStartConfig = attCurrentStartConfig;
}

void PaletteGik::cropMotion ( ChppGikRobotTask* inRobotTask )
{
    attLastMotion->clear();
    attLastMotion->appendMotion ( inRobotTask->solutionMotion() );

    double eT = inRobotTask->solutionMotion().endTime();

    bool assertion = inRobotTask->solutionMotion().configAtTime ( eT, attNextStartConfig );

    if ( !assertion )
        std::cout << "Debug : PaletteGik::cropMotion(): error : happens if motion is empty\n";
}

bool PaletteGik::stepback()
{
    //the robot is static at the current configuration
    attStandingRobot->staticState ( attCurrentStartConfig );
    ChppGikStepBackTask stpbkTask ( attStandingRobot,attSamplingPeriod );

    stpbkTask.showResolutionTime ( true );
    bool solved = stpbkTask.solve();

    if ( solved )
        cropMotion ( &stpbkTask );

    return solved;
}

bool PaletteGik::lookat ( const vector3d& targetPoint )
{
    //the robot is static at the current configuration
    attStandingRobot->staticState ( attCurrentStartConfig );

    //Reset planner
    ChppGikWholeBodyTask* lookTask = new ChppGikWholeBodyTask ( attStandingRobot,attSamplingPeriod,1 );

    CjrlGikGazeConstraint *gc = gikFactory.createGazeConstraint ( *robot, targetPoint );
    unsigned int priority = 1;

    std::cout << targetPoint << "\n";

    lookTask->addStateConstraint ( gc,priority );

    lookTask->enableStepping ( false );
    lookTask->showResolutionTime ( true );

    bool solved = lookTask->solve();

    if ( solved )
        cropMotion ( lookTask );

    delete gc;
    delete lookTask;
    return solved;
}

/* --- PaletteGik::reach --------------------------------------------- */

/** Place one hand at the given position
 *
 * \return	True if a solution has been found
 */

bool
PaletteGik::reach ( bool taskIsForRightHand, bool doOrientation, bool doLook,
                    bool steppingEnabled, const vector3d& inReachTarget,
                    const vector3d& inTargetOrientation,
                    const vector3d& inGazeTarget )
{
    bool solved;

    attStandingRobot->staticState ( attCurrentStartConfig );

    ChppGikReachTask reachTask ( attStandingRobot, attSamplingPeriod );

    reachTask.target ( inReachTarget, taskIsForRightHand );
    reachTask.steppingEnabled ( steppingEnabled );

    if ( doLook )
        reachTask.gazeTarget ( inGazeTarget );
    else
        reachTask.disableGaze();

    if ( doOrientation )
        reachTask.handOrientation ( inTargetOrientation );
    else
        reachTask.disableOrientation();

    reachTask.showResolutionTime ( true );
    solved = reachTask.solve();

    if ( solved )
        cropMotion ( &reachTask );
    return solved;
}


/* --- PaletteGik::grasp --------------------------------------------- */

/** Generate a simple grasping motion at the given 3D point
 *
 * \return	True if a solution has been found
 */

bool
PaletteGik::grasp ( bool taskIsForRightHand, const vector3d &targetPoint,
                    double clench )
{
    bool solved;
    matrix4d t3d;
    vector3d back;

    attStandingRobot->staticState ( attCurrentStartConfig );

    ChppGikHandTask handopenTask ( attStandingRobot, attSamplingPeriod );
    handopenTask.forRightHand ( taskIsForRightHand );
    handopenTask.targetClench ( 0.0 );

    ChppGikHandTask handcloseTask ( attStandingRobot, attSamplingPeriod );
    handcloseTask.forRightHand ( taskIsForRightHand );
    handcloseTask.targetClench ( clench );

    ChppGikReachTask reachTask ( attStandingRobot, attSamplingPeriod );
    reachTask.target ( targetPoint, taskIsForRightHand );
    reachTask.gazeTarget ( targetPoint );
    reachTask.disableOrientation();
    reachTask.steppingEnabled ( false );

    ChppGikReachTask backTask ( attStandingRobot, attSamplingPeriod );
    if ( taskIsForRightHand )
    {
        t3d =  robot->rightWrist()->currentTransformation();
        back = t3d * robot->rightHand()->centerInWristFrame();
    }
    else
    {
        t3d = robot->leftWrist()->currentTransformation();
        back = t3d * robot->leftHand()->centerInWristFrame();
    }
    backTask.target ( back, taskIsForRightHand );
    backTask.gazeTarget ( back );


    ChppGikMetaTask taskStack ( attStandingRobot, attSamplingPeriod );
    taskStack.pushbackTask ( &handopenTask );
    taskStack.pushbackTask ( &reachTask );
    taskStack.pushbackTask ( &handcloseTask );
    taskStack.pushbackTask ( &backTask );

    taskStack.showResolutionTime ( true );
    solved = taskStack.solve();

    if ( solved )
        cropMotion ( &taskStack );

    return solved;
}


/* --- PaletteGik::grasp_via ----------------------------------------- */

/** Generate a simple grasping motion at the given 3D point
 *
 * \return	True if a solution has been found
 */

bool
PaletteGik::grasp_via ( bool taskIsForRightHand, const vector3d &viaPoint,
                        const vector3d &targetPoint, double clench )
{
    bool solved;
    matrix4d t3d;
    vector3d back;

    attStandingRobot->staticState ( attCurrentStartConfig );

    ChppGikHandTask handopenTask ( attStandingRobot, attSamplingPeriod );
    handopenTask.forRightHand ( taskIsForRightHand );
    handopenTask.targetClench ( 0.0 );

    ChppGikHandTask handcloseTask ( attStandingRobot, attSamplingPeriod );
    handcloseTask.forRightHand ( taskIsForRightHand );
    handcloseTask.targetClench ( clench );

    ChppGikReachTask viaTask ( attStandingRobot, attSamplingPeriod );
    viaTask.target ( viaPoint, taskIsForRightHand );
    viaTask.gazeTarget ( viaPoint );
    viaTask.disableOrientation();
    viaTask.steppingEnabled ( true );

    ChppGikReachTask reachTask ( attStandingRobot, attSamplingPeriod );
    reachTask.target ( targetPoint, taskIsForRightHand );
    reachTask.gazeTarget ( targetPoint );
    reachTask.disableOrientation();
    reachTask.steppingEnabled ( false );

    ChppGikReachTask outTask ( attStandingRobot, attSamplingPeriod );
    outTask.target ( viaPoint, taskIsForRightHand );
    outTask.gazeTarget ( viaPoint );
    outTask.disableOrientation();
    outTask.steppingEnabled ( false );

    ChppGikHalfSittingTask backTask =
        ChppGikHalfSittingTask ( attStandingRobot, attSamplingPeriod );


    ChppGikMetaTask taskStack ( attStandingRobot, attSamplingPeriod );
    taskStack.pushbackTask ( &viaTask );
    taskStack.pushbackTask ( &handopenTask );
    taskStack.pushbackTask ( &reachTask );
    taskStack.pushbackTask ( &handcloseTask );
    taskStack.pushbackTask ( &outTask );
    taskStack.pushbackTask ( &backTask );

    taskStack.showResolutionTime ( true );
    solved = taskStack.solve();

    if ( solved )
        cropMotion ( &taskStack );

    return solved;
}


/* --- PaletteGik::release ------------------------------------------- */

/** Generate a simple release motion at the given 3D point
 *
 * \return	True if a solution has been found
 */

bool
PaletteGik::release ( bool taskIsForRightHand, const vector3d &targetPoint,
                      double clench )
{
    bool solved;
    matrix4d t3d;
    vector3d back;

    attStandingRobot->staticState ( attCurrentStartConfig );

    ChppGikReachTask reachTask ( attStandingRobot, attSamplingPeriod );
    reachTask.target ( targetPoint, taskIsForRightHand );
    reachTask.gazeTarget ( targetPoint );
    reachTask.disableOrientation();
    reachTask.steppingEnabled ( false );

    ChppGikHandTask handopenTask ( attStandingRobot, attSamplingPeriod );
    handopenTask.forRightHand ( taskIsForRightHand );
    handopenTask.targetClench ( clench );

    ChppGikReachTask backTask ( attStandingRobot, attSamplingPeriod );
    if ( taskIsForRightHand )
    {
        t3d =  robot->rightWrist()->currentTransformation();
        back = t3d * robot->rightHand()->centerInWristFrame();
    }
    else
    {
        t3d = robot->leftWrist()->currentTransformation();
        back = t3d * robot->leftHand()->centerInWristFrame();
    }
    backTask.target ( back, taskIsForRightHand );
    backTask.gazeTarget ( back );


    ChppGikMetaTask taskStack ( attStandingRobot, attSamplingPeriod );
    taskStack.pushbackTask ( &reachTask );
    taskStack.pushbackTask ( &handopenTask );
    taskStack.pushbackTask ( &backTask );

    taskStack.showResolutionTime ( true );
    solved = taskStack.solve();

    if ( solved )
        cropMotion ( &taskStack );

    return solved;
}


/* --- PaletteGik::release_via --------------------------------------- */

/** Generate a simple release motion at the given 3D point
 *
 * \return	True if a solution has been found
 */

bool
PaletteGik::release_via ( bool taskIsForRightHand, const vector3d &viaPoint,
                          const vector3d &targetPoint, double clench )
{
    bool solved;
    matrix4d t3d;
    vector3d back;

    attStandingRobot->staticState ( attCurrentStartConfig );

    ChppGikReachTask viaTask ( attStandingRobot, attSamplingPeriod );
    viaTask.target ( viaPoint, taskIsForRightHand );
    viaTask.gazeTarget ( targetPoint );
    viaTask.disableOrientation();
    viaTask.steppingEnabled ( true );

    ChppGikReachTask reachTask ( attStandingRobot, attSamplingPeriod );
    reachTask.target ( targetPoint, taskIsForRightHand );
    reachTask.gazeTarget ( targetPoint );
    reachTask.disableOrientation();
    reachTask.steppingEnabled ( false );

    ChppGikReachTask outTask ( attStandingRobot, attSamplingPeriod );
    outTask.target ( viaPoint, taskIsForRightHand );
    outTask.gazeTarget ( targetPoint );
    outTask.disableOrientation();
    outTask.steppingEnabled ( false );

    ChppGikHandTask handopenTask ( attStandingRobot, attSamplingPeriod );
    handopenTask.forRightHand ( taskIsForRightHand );
    handopenTask.targetClench ( clench );

    ChppGikHalfSittingTask backTask =
        ChppGikHalfSittingTask ( attStandingRobot, attSamplingPeriod );


    ChppGikMetaTask taskStack ( attStandingRobot, attSamplingPeriod );
    taskStack.pushbackTask ( &viaTask );
    taskStack.pushbackTask ( &reachTask );
    taskStack.pushbackTask ( &handopenTask );
    taskStack.pushbackTask ( &outTask );
    taskStack.pushbackTask ( &backTask );

    taskStack.showResolutionTime ( true );
    solved = taskStack.solve();

    if ( solved )
        cropMotion ( &taskStack );

    return solved;
}


/* --- PaletteGik::halfsitting --------------------------------------------- */

/** Go to the (initial) half-sitting stance
 *
 * \return	True if a solution has been found
 */

bool PaletteGik::halfsitting()
{
    bool solved;

    attStandingRobot->staticState ( attCurrentStartConfig );

    ChppGikHalfSittingTask halfsittingTask ( attStandingRobot, attSamplingPeriod );
    halfsittingTask.showResolutionTime ( true );

    solved = halfsittingTask.solve();
    if ( solved )
        cropMotion ( &halfsittingTask );

    return solved;
}


bool PaletteGik::handGrasp ( bool taskIsForRightHand, double valTighten )
{
    //the robot is static at the current configuration
    attStandingRobot->staticState ( attCurrentStartConfig );

    ChppGikHandTask* handTask = new ChppGikHandTask ( attStandingRobot,attSamplingPeriod );
    handTask->forRightHand ( taskIsForRightHand );
    handTask->targetClench ( valTighten );

    handTask->showResolutionTime ( true );
    bool solved = handTask->solve();

    if ( solved )
        cropMotion ( handTask );

    delete handTask;
    return solved;
}

void PaletteGik::apply()
{
    attCurrentStartConfig = attNextStartConfig;
}


/* --- PaletteGik::firstSample --------------------------------------- */

/** Return first motion sample from the last computed motion.
 *
 * \return true on success
 */

bool
PaletteGik::firstSample ( vectorN &config, vector3d &ZMPwstObs, double &timeRef )
{
    const ChppRobotMotionSample *s = attLastMotion->firstSample();

    if ( !s )
        return false;

    config = s->configuration;
    ZMPwstObs = s->ZMPwstObs;
    attLastSampleTime = timeRef = 0.0;
    return true;
}


/* --- PaletteGik::nextSample ---------------------------------------- */

/** Return next motion sample from the last computed motion.
 *
 * \return true on success
 */

bool
PaletteGik::nextSample ( vectorN &config, vector3d &ZMPwstObs, double &timeRef )
{
    const ChppRobotMotionSample *s = attLastMotion->nextSample();
    attLastSampleTime += attSamplingPeriod;

    if ( !s )
        return false;

    config = s->configuration;
    ZMPwstObs = s->ZMPwstObs;
    timeRef = attLastSampleTime;
    return true;
}

void PaletteGik::dumpLastMotion()
{
    attLastMotion->dumpTo ( attHRP2MotionName.c_str() );
}









/*********************************************************************************************************/
/*********************************************************************************************************/

void PaletteGik::cropGik()
{
    attLastMotion->clear();

    if ( attGikMotion->empty() )
    {
        std::cout << "PaletteGik::cropGik(): motion was empty" << std::endl;
        return;
    }

    attLastMotion->appendMotion ( *attGikMotion );
    attNextStartConfig = attGikMotion->lastSample()->configuration;
}

bool PaletteGik::elementLookat ( const vector3d& targetPoint )
{
    ChppGikVectorizableConstraint *gc = new ChppGikGazeConstraint ( *robot, targetPoint );
    std::vector<ChppGikVectorizableConstraint*> tasks;
    tasks.push_back ( gc );
    bool solved = solveTaskStack ( tasks );
    delete gc;
    return solved;
}


bool PaletteGik::elementReach ( bool taskIsForRightHand, bool doOrientation, bool doLook,
                                const vector3d& inReachTarget,
                                const vector3d& inTargetOrientation,
                                const vector3d& inGazeTarget )
{
    bool solved;

    std::vector<ChppGikVectorizableConstraint*> tasks;
    CjrlJoint* joint;
    vector3d lpoint;
    CjrlHand* hand;
    if ( taskIsForRightHand )
    {
        joint = attStandingRobot->robot()->rightWrist();
        hand =  attStandingRobot->robot()->rightHand();

    }
    else
    {
        joint = attStandingRobot->robot()->leftWrist();
        hand =  attStandingRobot->robot()->leftHand();
    }
    lpoint = hand->centerInWristFrame();

    tasks.push_back ( new ChppGikPositionConstraint ( * ( attStandingRobot->robot() ),*joint,lpoint,inReachTarget ) );

    if ( doOrientation )
    {
        vector3d laxis = hand->okayAxisInWristFrame();
        tasks.push_back ( new ChppGikParallelConstraint ( * ( attStandingRobot->robot() ),*joint,laxis,inTargetOrientation ) );
    }
    if ( doLook )
        tasks.push_back ( new ChppGikGazeConstraint ( * ( attStandingRobot->robot() ), inGazeTarget ) );

    solved = solveTaskStack ( tasks );

    for ( unsigned int i=0;i<tasks.size();i++ )
        delete tasks[i];

    attGikMotion->dumpTo ( "elementReachMotion" );
    return solved;
}


bool PaletteGik::elementClench ( bool taskIsForRightHand, double clench )
{
    vectorN backupConfig = attStandingRobot->robot()->currentConfiguration();

    vector3d dummy;
    double clenchDuration = 2.0;
    CjrlHand* hand;
    if ( taskIsForRightHand )
        hand =  attStandingRobot->robot()->rightHand();
    else
        hand =  attStandingRobot->robot()->leftHand();

    unsigned int nSamples = ChppGikTools::timetoRank ( 0.0,clenchDuration, attSamplingPeriod ) +1;
    vectorN interpolationVector ( nSamples );
    bool contn = ChppGikTools::minJerkCurve ( clenchDuration,attSamplingPeriod,attStandingRobot->robot()->getHandClench ( hand ) ,0.0,0.0,clench,interpolationVector );

    if ( !contn )
        return false;

    attGikMotion->clear();
    double val;
    for ( unsigned int i=0; i<nSamples; i++ )
    {

        val = interpolationVector ( i );
        if ( val > 1-1e-5 )
            val = 1-1e-5;
        else
            if ( val < 1e-5 )
                val = 1e-5;
        attStandingRobot->robot()->setHandClench ( hand, val );
        attGikMotion->appendSample ( attStandingRobot->robot()->currentConfiguration(), dummy, dummy, dummy, dummy );
    }

    attStandingRobot->staticState ( backupConfig );
    return true;
}

bool PaletteGik::sittingLookat ( const vector3d& targetPoint )
{
    attStandingRobot->staticState ( attCurrentStartConfig );

    bool solved = elementLookat ( targetPoint );
    if ( solved )
        cropGik();

    return solved;
}

bool PaletteGik::sittingReach ( bool taskIsForRightHand, bool doOrientation, bool doLook,
                                const vector3d& inReachTarget,
                                const vector3d& inTargetOrientation,
                                const vector3d& inGazeTarget )
{
    attStandingRobot->staticState ( attCurrentStartConfig );

    bool solved = elementReach ( taskIsForRightHand, doOrientation, doLook,
                                 inReachTarget,
                                 inTargetOrientation,
                                 inGazeTarget );
    if ( solved )
        cropGik();

    return solved;
}

bool PaletteGik::sittingHandGrasp ( bool taskIsForRightHand, double valTighten )
{
    attStandingRobot->staticState ( attCurrentStartConfig );

    bool solved = elementClench ( taskIsForRightHand,valTighten );

    if ( solved )
        cropGik();

    return solved;
}

bool PaletteGik::sittingManipulation ( bool isForGrasp,bool taskIsForRightHand, const std::vector<vector3d>& inPickingPath, const std::vector<vector3d>& inReturningPath, double clench )
{
    ChppRobotMotion motion ( attStandingRobot->robot(),0.0,attSamplingPeriod );

    bool solved;

    if ( isForGrasp )
    {
        solved = elementClench ( taskIsForRightHand,0.0 );
        if ( solved )
        {
            motion.appendMotion ( *attGikMotion );
            attStandingRobot->staticState ( motion.lastSample()->configuration );
        }
        else
        {
            attStandingRobot->staticState ( attCurrentStartConfig );
            return false;
        }
    }

    for ( unsigned int i = 0; i<inPickingPath.size();i++ )
    {
        solved = elementReach ( taskIsForRightHand, false, true,
                                inPickingPath[i],
                                vector3d(),
                                inPickingPath[inPickingPath.size()-1] );

        if ( solved )
        {
            motion.appendMotion ( *attGikMotion );
            attStandingRobot->staticState ( motion.lastSample()->configuration );
        }
        else
        {
            attStandingRobot->staticState ( attCurrentStartConfig );
            return false;
        }
    }

    if ( !isForGrasp )
    {
        solved = elementClench ( taskIsForRightHand,0.0 );
        if ( solved )
        {
            motion.appendMotion ( *attGikMotion );
            attStandingRobot->staticState ( motion.lastSample()->configuration );
        }
        else
        {
            attStandingRobot->staticState ( attCurrentStartConfig );
            return false;
        }
    }

    if ( isForGrasp )
    {
        solved = elementClench ( taskIsForRightHand,clench );
        if ( solved )
        {
            motion.appendMotion ( *attGikMotion );
            attStandingRobot->staticState ( motion.lastSample()->configuration );
        }
        else
        {
            attStandingRobot->staticState ( attCurrentStartConfig );
            return false;
        }
    }

    vector3d reposVision;
    reposVision[0] = 1;
    reposVision[1] = 0.0;
    reposVision[2] = 0.6;
    for ( unsigned int i = 0; i<inReturningPath.size();i++ )
    {
        solved = elementReach ( taskIsForRightHand, false, true,
                                inReturningPath[i],
                                vector3d(),
                                reposVision );
        if ( solved )
        {
            motion.appendMotion ( *attGikMotion );
            attStandingRobot->staticState ( motion.lastSample()->configuration );
        }
        else
        {
            attStandingRobot->staticState ( attCurrentStartConfig );
            return false;
        }
    }

    if ( !isForGrasp )
    {
        solved = elementClench ( taskIsForRightHand,clench );
        if ( solved )
        {
            motion.appendMotion ( *attGikMotion );
            attStandingRobot->staticState ( motion.lastSample()->configuration );
        }
        else
        {
            attStandingRobot->staticState ( attCurrentStartConfig );
            return false;
        }
    }

    attLastMotion->clear();
    attLastMotion->appendMotion ( motion );
    attNextStartConfig = attLastMotion->lastSample()->configuration;

    return true;
}


bool PaletteGik::solveTaskStack ( std::vector<ChppGikVectorizableConstraint*>& inTasks )
{
    attGikMotion->clear();
    vectorN backupConfig = attStandingRobot->robot()->currentConfiguration();
    double startTime = 0.0;

    std::vector<CjrlGikStateConstraint*> gikconstraints;
    std::vector<ChppGikInterpolatedElement*> Elements;
    std::vector<double> dampers;
    for ( unsigned int i = 0;i<inTasks.size();i++ )
    {
        Elements.push_back ( new ChppGikInterpolatedElement ( attStandingRobot->robot(),inTasks[i],i,startTime,attSittingTasksDuration,attSamplingPeriod ) );
        dampers.push_back ( 0.8 );
    }

    double time = startTime+attSamplingPeriod;
    double endtime = startTime+attSittingTasksDuration+attSamplingPeriod/2;
    bool gikStepSuccess = true;
    vector3d dummy;

    while ( time < endtime )
    {
        gikconstraints.clear();
        for ( unsigned int i = 0;i<inTasks.size();i++ )
            gikconstraints.push_back ( Elements[i]->stateConstraintAtTime ( time ) );

        gikStepSuccess = solveConstraintStack ( gikconstraints, dampers );

        if ( gikStepSuccess )
            attGikMotion->appendSample ( attStandingRobot->robot()->currentConfiguration(), dummy,dummy,dummy,dummy );
        else
            break;
        time += attSamplingPeriod;
    }

    for ( unsigned int i = 0; i<Elements.size(); i++ )
        delete Elements[i];


    for ( unsigned int i = 0; i<inTasks.size();i++ )
    {
        inTasks[i]->computeValue();
        inTasks[i]->value();
        std::cout << inTasks[i]->value() << std::endl;
    }

    attStandingRobot->staticState ( backupConfig );


    return gikStepSuccess;
}

bool PaletteGik::solveConstraintStack ( std::vector<CjrlGikStateConstraint*>& inConstraints, const std::vector<double>& inDamp )
{
    std::vector<double> prevVals, curVals;
    std::vector<double> dampers = inDamp;

    attBackupConfig = attStandingRobot->robot()->currentConfiguration();
    attBackupVelocity = attStandingRobot->robot()->currentVelocity();


    attGikSolver->prepare ( inConstraints );

    for ( unsigned int i = 0; i<inConstraints.size();i++ )
    {
        prevVals.push_back ( norm_2 ( inConstraints[i]->value() ) );
        curVals.push_back ( prevVals[i] );
    }

    //attGikSolver->solve( inConstraints );
    attGikSolver->solve ( inConstraints, dampers );
    attStandingRobot->updateRobot ( attGikSolver->solutionRootPose(), attGikSolver->solutionJointConfiguration(), attSamplingPeriod );
    for ( unsigned int i = 0; i<inConstraints.size();i++ )
    {
        inConstraints[i]->computeValue();
        inConstraints[i]->value();
        curVals[i] = norm_2 ( inConstraints[i]->value() );
    }

    if ( curVals[0] > prevVals[0] )
    {
        attStandingRobot->robot()->currentConfiguration ( attBackupConfig );
        attStandingRobot->robot()->currentVelocity ( attBackupVelocity );
        attStandingRobot->robot()->computeForwardKinematics();
        return false;
    }

    return true;
}

